/*****************************************************************************/
#ifdef COMMENTS_WITH_COMMENTS
/*
                                   Login.c


    THE GNU GENERAL PUBLIC LICENSE APPLIES DOUBLY TO ANYTHING TO DO WITH
                    AUTHENTICATION AND AUTHORIZATION!

    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License, or any later
    version.

>   This package is distributed in the hope that it will be useful,
>   but WITHOUT ANY WARRANTY; without even the implied warranty of
>   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
>   GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


DESCRIPTION
-----------
This module provides the core of a authorization management mechanism.

This module has NOTHING TO DO with REMOTE_USER source where the supporting web
server handles all authentication!

Obviously propagating sensitive data such as passwords around over a network
has it's dangers.  With normal HTTP authorization fields this happens all the
time (the credentials are base-64 encoded but not encrypted making them trivial
to discover).  This is why it is best for such transactions to occur using SSL. 
Certainly with SYSUAF based authentication you would imagine this to be
mandatory.  So it shouldn't be a real issue as far as the transaction goes but
such information also shouldn't be trivially recovered from either the HTML
or other browser source, or from cookie content.  VWcms only uses the
credentials (username/password combination) once during login and then supplies
an encrypted assertion that they have been presented.

The encryption is performed using the RC4 algorithm and a minimum 192 bit key
(24  characters) supplied from site configuration.  This makes it non-trivial
to recover such data.  Also, when VWCMS is logged-out the content of those
credentials are removed from the browser.

Sessions are also tracked for idle time.  If more than this period occurs
between successive requests VWCMS will present the login page and require
reauthentication.
 

RC4 ENCRYPTION
--------------
I am certainly not a cryptography specialist so be sure to take anything (and
everything) I say here with a large pinch of salt!

RC4 is a very commonly deployed, efficient and arguably safe-enough encryption
scheme.

  http://en.wikipedia.org/wiki/Rc4

While not necessarily authoritative the above reference states:

  "In cryptography, RC4 (also known as ARC4 or ARCFOUR) is the most
   widely-used software stream cipher and is used in popular protocols
   such as Secure Sockets Layer (SSL) (to protect Internet traffic) ...
   While remarkable in its simplicity, RC4 falls short of the high
   standards of security set by cryptographers ... However, some systems
   based on RC4 are secure enough for practical use."

And into that latter category VWCMS (hopefully) falls.  Keep your browser
secure while in use and [logout] to clear the actual encrypted content when
appropriate.

To assist in keeping the credential encryption secure ensure the VWCMS_CONFIG
file is protected for general access and ensure the [login-secret] string used
for the encryption key contains a whole range of characters.  VWCMS insists
this key is at least 24 characters in length.  An *example* key might be

  [login-secret]  aic84+-[QsfOPa,dk;'12ndjve64_90mhd43m*2$

The secret can be modified at any time but consider that client sessions
currently in-progress will require to re-login the next time they access
VWCMS.

For implementation detail see module RC4.C

The encrypted data is additionally base64-encoded to make it suitable for
character-oriented schemes such as HTTP cookies.


CREDENTIALS COOKIE
------------------
For all their issues and the disdain in which they are held
by some they are an effective and transparent mechanism for providing state
information of all sorts with HTTP requests and so have become the chosen
mechanism for VWCMS autogenous authentication.  Don't like them?  Well don't
manage a VWcms site!

The cookie value is a base64-encoded, RC4-encrypted series of null-terminated
ASCII strings.  Some strings represent numbers, others text.  When a cookie
value is generated the data strings are preceded by a variable-length, number. 
This serves to add some quasi-indeterminate data to the mix and results in
startlingly different encrypted/encoded cookie values each time one is
generated (and is therefore perhaps of benefit to encryption security through
an obfuscation factor).

Following the obfuscation factor there is the U**x time in seconds as a deciaml
number, after which the session is considered idle.  Following that another
decimal number, the U**x time in seconds.  This is the point after which a
session is considered idle.  Following these are three strings; the IP address
of the client (from CGI variable REMOTE_ADDR), the authenticated username, and
the original encryption key string, used to validate the assertion.

A credentials cookie looks something like, plain and encrypted/encoded:

  \256\01160471530\0192.168.1.2\0DANIEL\0somethinglikeanrc4key\0
  SZX9Lf8uefaXl3fsAzazonP0mV6pTa6hWg26nb/EnE8rIxdH4Lxj1KBWfOVZtrdQnnBUisZmP

(where \0 is a null-terminator)

It should be emphasised that the credential data is transmitted and stored
encrypted.  It is only plain data when being handled internally.


LOGIN ALIAS
-----------
For autogenous authentication the configuration directive [login-alias] allows
the user-supplied 'username' strings to be mapped to other strings before being
authenticated.  An example use might be using the user email address, or local
mailbox name, as the 'username' and mapping that to a VMS username before
authenticating against the SYSUAF.  The alias mapping is performed using a
case-insensitive match.  The format is one alias mapping per line.

  [login-alias]
  Mark.Daniel/DANIEL
  Fred.Bloggs/BLOGGS

If a match is not made and no mapping has occurred the default (drop-through)
behaviour is to use the original user-supplied 'username' string and so in the
above configuration both "Mark.Daniel" and "DANIEL" would authenticate against
the DANIEL account, "Fred.Bloggs" and "BLOGGS" against BLOGGS.  Of course any
other supplied username could also authenticate directly using the username. 
To prevent any other than an aliased username being used for authentication
conclude the list with "*/-".

  [login-alias]
  Mark.Daniel/DANIEL
  Fred.Bloggs/BLOGGS
  */-

In the above example only "Mark.Daniel" and "Freg.Bloggs" can be supplied to
allow authentication.  The (somewhat redundant) mapping "*/*" will explicitly
direct authentication to occur using the supplied 'username' (the default
drop-through behaviour anyway).


EXTERNAL AGENT AUTHENTICATION
-----------------------------
If the [login-source] begins with a '@' character then it activates an external
authentication DCL procedure within a subprocess.  The DCL procedure must be
specified following the '@' symbol and must have execute permission for the
scripting account.  For example:

  [login-source] @cgi-bin:[000000]vwcms_authenticate.com

Optional parameters can be provided on the configuration directive
(command-line):

  [login-source] @cgi-bin:[000000]vwcms_authenticate.com "param1" "PARAM2"

When activated the procedure has a further three parameters appended to any
supplied in the above configuration directive; "<username>", "<password>" and
"<remote_addr>", to be used in the authentication processing.

The procedure should exit with any success status (e.g. SS$_NORMAL) to indicate
successful authentication or any error status (e.g. SS$_NOPRIV) to indicate
authentication failure.

This is a (very simple) example.

  $ SET ON
  $ OK = 1
  $ NOPRIV = 36
  $ P1 = F$EDIT(P1,"UPCASE")
  $ IF P1 .EQS. "CURLY" .AND. P2 .EQS. "JeromE" THEN EXIT OK
  $ IF P1 .EQS. "LARRY" .AND. P2 .EQS. "lOUIs" THEN EXIT OK
  $ IF P1 .EQS. "MOE" .AND. P2 .EQS. "harrY" THEN EXIT OK
  $ EXIT NOPRIV

This would probably be more like a working one.

  $ SET ON
  $ VWAUTH = "$location:VWAUTH.EXE"
  $ VWAUTH "''P1'" "''P2'" "''P3'"
  $ EXIT $STATUS


COPYRIGHT
---------
Copyright (C) 2008-2013 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
25-MAR-2009  MGD  LoginAlias() to map user-supplied login name to user name
24-MAR-2008  MGD  initial
*/
#endif /* COMMENTS_WITH_COMMENTS */
/*****************************************************************************/

#ifdef WASD_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

/* VMS related header files */
#include <cgilib.h>
#include <descrip.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>

#define BOOL int
#define TRUE 1
#define FALSE 0

/* application related header files */
#include "authvms.h"
#include "config.h"
#include "content.h"
#include "cookie.h"
#include "login.h"
#include "rc4.h"
#include "other.h"
#include "template.h"
#include "vwcms.h"

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

/* version 1 of the cookie structure */
#define LOGIN_COOKIE_VERSION 1

#define FI_LI "LOGIN", __LINE__

/******************/
/* global storage */
/******************/

char  LoginUserName [LOGIN_USERNAME_SIZE+1];

/********************/
/* external storage */
/********************/

extern BOOL  Debug,
             WatchEnabled;

extern char  *CgiHttpRefererPtr,
             *CgiPathInfoPtr; 

extern CONFIG_DATA  VWcmsConfig;

/***********************/
/* required prototypes */
/***********************/

int ErrorExit (int, char*, int);

int lib$add_times (__unknown_params);
int lib$sub_times (__unknown_params);
int lib$spawn (__unknown_params);
int sys$fao (__unknown_params);
int sys$gettim (__unknown_params);

/*****************************************************************************/
/*
Configure login.
*/

void LoginConfig ()

{
   unsigned char  *ucptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginConfig()\n");

   if (!VWcmsConfig.LoginSource) return;

   if (strsame (VWcmsConfig.LoginSource, "REMOTE_USER", -1))
   {
      /* when the web server is doing it all do no more */
      VWcmsConfig.LoginEnabled = VWcmsConfig.LoginRemoteUser = TRUE;
      return;
   }

   if (strsame (VWcmsConfig.LoginSource, "VMS", -1))
      VWcmsConfig.LoginVms = TRUE;
   else
   if (strsame (VWcmsConfig.LoginSource, "VWCMS", -1))
      VWcmsConfig.LoginVWcms = TRUE;
   else
   if (*VWcmsConfig.LoginSource == '@')
      VWcmsConfig.LoginAgent = TRUE;
   else
   {
      TemplateAdvise ("ECONFIG login");
      if (WatchEnabled)
         WatchThis (FI_LI, "LOGIN source unknown !AZ",
                    VWcmsConfig.LoginSource);
      return;
   }

   /* if not configured it'll certainly be less than the minimum length! */
   if (!VWcmsConfig.LoginSecret) VWcmsConfig.LoginSecret = "";

   /* make it a little less like ASCII data by swapping the nybles */
   for (ucptr = (unsigned char*)VWcmsConfig.LoginSecret; *ucptr; ucptr++)
      *ucptr = ((*ucptr & 0xf0) >> 4) | ((*ucptr & 0x0f) << 4);

   /* check that it's a large enough secret */
   if (ucptr - (unsigned char*)VWcmsConfig.LoginSecret < 24)
   {
      TemplateAdvise ("ECONFIG login");
      if (WatchEnabled)
         WatchThis (FI_LI, "LOGIN secret < 24 characters");
      return;
   }

   /* finally it's true! */
   VWcmsConfig.LoginEnabled = TRUE;

   if (!VWcmsConfig.LoginIdleSeconds)
      VWcmsConfig.LoginIdleSeconds = LOGIN_IDLE_SECONDS_DEFAULT;
}

/*****************************************************************************/
/*
Returns true if login cannot be done under these circumstances.
Currently; restriction from logging in via clear-text HTTP.
*/

BOOL LoginRestricted ()

{
   BOOL  restr;
   char  *cptr;


   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginRestricted()\n");

   if (!VWcmsConfig.LoginEnabled) return (FALSE);

   if (VWcmsConfig.LoginSSL)
   {
      /* login only via SSL (try and cater for all server environments) */
      restr = TRUE;
      cptr = GetCgiVar ("REQUEST_SCHEME"); 
      if (!strcmp (cptr, "https:")) restr = FALSE;
      if (restr) restr = !(BOOL)GetCgiVarNull ("SSL_PROTOCOL");
      if (restr) restr = !(BOOL)GetCgiVarNull ("SSL_VERSION");
      if (restr)
      {
         cptr = GetCgiVar ("SERVER_PORT"); 
         if (!strcmp (cptr, "443")) restr = FALSE;
      }
      if (restr)
      {
         /* login is not via SSL */
         if (strsame (VWcmsConfig.LoginSSL, "https://", 8))
            CgiLibResponseRedirect ("https://%s%s",
               GetCgiVar("SERVER_NAME"), GetCgiVar("REQUEST_URI"));
         else
            CgiLibResponseError (FI_LI, 0, VWcmsConfig.LoginSSL);
      }
   }
   else
      restr = FALSE;

   if (WatchEnabled)
      WatchThis (FI_LI, "LOGIN restricted !AZ", restr ? "TRUE" : "FALSE");

   return (restr);
}

/*****************************************************************************/
/*
Primary application interface to verify the user credentials whether from the
login page (initial) or the session cookie.  Set the 'LoginUserName' global
storage and return true if credentials valid, reset and return false if
invalid.
*/ 

int LoginVerifyUser ()

{
   int  CookieValueLength,
        IdleSeconds,
        VerifySeconds;
   unsigned long  TimeSeconds;                       
   char  *cptr, *czptr, *sptr, *zptr,
         *CookieValuePtr;
   char  IdleString [64],
         RemoteAddr [128],
         SharedSecret [256],
         /* plenty of space as AuthVmsVerifyUser() can map a username into it */
         UserName [64];         

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginVerifyUser()\n");

   if (!VWcmsConfig.LoginEnabled) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (VWcmsConfig.LoginRemoteUser) ErrorExit (SS$_BUGCHECK, FI_LI);

   LoginUserName[0] = '\0';

   CookieValuePtr = LoginGetSessionCookie (&CookieValueLength);

   if (!CookieValuePtr || !CookieValueLength)
   {
      /******************/
      /* no credentials */
      /******************/

      return (LOGIN_VERIFY_NONE);
   }

   /******************/
   /* session cookie */
   /******************/

   czptr = (cptr = CookieValuePtr) + CookieValueLength;

   /* skip across the obfuscation factor */
   while (*cptr && cptr < czptr) cptr++;
   if (cptr >= czptr) return (LOGIN_VERIFY_FAIL); 
   cptr++;

   /* idle seconds string */
   IdleSeconds = atoi(cptr);
   while (*cptr && cptr < czptr) cptr++;
   if (cptr >= czptr) return (LOGIN_VERIFY_FAIL);
   cptr++;

   /* remote IP address */
   zptr = (sptr = RemoteAddr) + sizeof(RemoteAddr)-1;
   while (*cptr && cptr < czptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr) return (LOGIN_VERIFY_FAIL);
   *sptr = '\0';
   while (*cptr && cptr < czptr) cptr++;
   if (cptr >= czptr) return (LOGIN_VERIFY_FAIL);
   cptr++;

   /* username (or its alias) string */
   zptr = (sptr = UserName) + sizeof(UserName)-1;
   while (*cptr && cptr < czptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr) return (LOGIN_VERIFY_FAIL);
   *sptr = '\0';
   while (*cptr && cptr < czptr) cptr++;
   if (cptr >= czptr) return (LOGIN_VERIFY_FAIL);
   cptr++;

   /* shared secret */
   zptr = (sptr = SharedSecret) + sizeof(SharedSecret)-1;
   while (*cptr && cptr < czptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr) return (LOGIN_VERIFY_FAIL);
   *sptr = '\0';

   if (strcmp (SharedSecret, VWcmsConfig.LoginSecret))
   {
      if (WatchEnabled) WatchThis (FI_LI, "LOGIN corrupt shared secret");
      return (LOGIN_VERIFY_FAIL);
   }

   time (&TimeSeconds);

   if (WatchEnabled)
      WatchThis (FI_LI, "LOGIN !AZ !AZ !UL (!SL)",
                 RemoteAddr, UserName, IdleSeconds, IdleSeconds-TimeSeconds);

   if (IdleSeconds < TimeSeconds)
   {
      /****************/
      /* idle session */
      /****************/

      /* cause re-login */
      CookieSet (LOGIN_SESSION_COOKIE, NULL);
      return (LOGIN_VERIFY_IDLE);
   }

   /* reset the the time values in the session cookie */
   LoginSetSession (UserName, RemoteAddr);

   /* copy the non-validated username into the global buffer */
   zptr = (sptr = LoginUserName) + sizeof(LoginUserName)-1;
   for (cptr = UserName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr > zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   *sptr = '\0';

   return (LOGIN_VERIFY_OK);
}

/*****************************************************************************/
/*
From the login page HTML form fields.
*/

BOOL LoginInitial ()

{
   BOOL  ok;
   char  *cptr,
         *Base64Ptr,
         *UserName,
         *Password,
         *RemoteAddr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginInitial()\n");

   if (!VWcmsConfig.LoginEnabled) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (VWcmsConfig.LoginRemoteUser) ErrorExit (SS$_BUGCHECK, FI_LI);

   LoginUserName[0] = '\0';

   UserName = GetCgiVarNull ("FORM_USER");
   Password = GetCgiVarNull ("FORM_PASS");
   RemoteAddr = GetCgiVarNull ("REMOTE_ADDR");

   if (!RemoteAddr || !*RemoteAddr) ErrorExit (SS$_BUGCHECK, FI_LI);

   if ((!UserName || !*UserName) &&
       (!Password || !*Password))
   {
      if (WatchEnabled) WatchThis (FI_LI, "LOGIN abandoned");
      CookieSet (LOGIN_SESSION_COOKIE, NULL);
      return (TRUE);
   }

   if (!UserName || !*UserName ||
       !Password || !*Password)
   {
      if (WatchEnabled) WatchThis (FI_LI, "LOGIN username/password incomplete");
      ok = FALSE;
   }
   else
      ok = LoginAuthUser (UserName, Password, RemoteAddr);

   return (ok);
}

/*****************************************************************************/
/*
Authenticate the user credentials against the configured authentication source. 
If valid set the session cookie and return true.  Return false if invalid.
Includes a delay on invalid credentials to slow down repeated attempts.
*/

BOOL LoginAuthUser
(
char *UserName,
char *Password,
char *RemoteAddr
)
{
   BOOL  ok = FALSE;
   int  status;
   char  *cptr, *sptr, *zptr,
         *UserAlias;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginAuthUser() |%s|\n", UserName);

   if (!VWcmsConfig.LoginEnabled) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (VWcmsConfig.LoginRemoteUser) ErrorExit (SS$_BUGCHECK, FI_LI);

   LoginUserName[0] = '\0';

   if (!LoginAlias (&UserName, &UserAlias)) return (FALSE);

   if (VWcmsConfig.LoginVms)
   {
      status = AuthVmsVerifyUser (UserName, Password, RemoteAddr);
      if (VMSnok (status))
      {
         sleep (2);
         return (FALSE);
      }
   }
   else
   if (VWcmsConfig.LoginVWcms)
   {
      status = ConfigVerifyUser  (UserName, Password, RemoteAddr);
      if (VMSnok (status))
      {
         sleep (2);
         return (FALSE);
      }
   }
   else
   if (VWcmsConfig.LoginAgent)
   {
      status = LoginAgent (UserName, Password, RemoteAddr);
      if (VMSnok (status))
      {
         sleep (2);
         return (FALSE);
      }
   }
   else
       ErrorExit (SS$_BUGCHECK, FI_LI);

   zptr = (sptr = LoginUserName) + sizeof(LoginUserName)-1;
   if (UserAlias && UserAlias[0])
      for (cptr = UserAlias; *cptr && sptr < zptr; *sptr++ = *cptr++);
   else
      for (cptr = UserName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr > zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   *sptr = '\0';

   LoginSetSession (LoginUserName, RemoteAddr);

   return (TRUE);
}

/*****************************************************************************/
#ifdef COMMENTS_WITH_COMMENTS
/*
Map an alias login name to a user name to be used for authentication.  Modify
the alias name pointer to an empty string if no mapping was found.  Return TRUE
when the user name pointer is valid for authentication (mapped or not).  If the
directive "*/-" is encountered return FALSE to indicate that no match means no
authentication.  If the directive "*/*" return TRUE.
*/ 
#endif /* COMMENTS_WITH_COMMENTS */

BOOL LoginAlias
(
char **UserNamePtrPtr,
char **AliasNamePtrPtr
)
{
   static char  AliasName [128],
                UserName [128];

   char  *aptr, *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginAlias()\n");

   /* an empty string indicates there was no mapping */
   *AliasNamePtrPtr = "";

   if (!(aptr = VWcmsConfig.LoginAlias)) return (TRUE);

   while (*aptr)
   {
      while (*aptr && isspace(*aptr)) aptr++;
      if (!*aptr) break;
      if (SAME2(aptr,'*/'))
      {
         if (WatchEnabled) WatchThis (FI_LI, "ALIAS !3AZ", aptr);
         if (*(aptr+2) == '*') return (TRUE);
         return (FALSE);
      }
      sptr = *UserNamePtrPtr;
      cptr = aptr;
      while (*aptr && *aptr != '/' && *sptr &&
             tolower(*aptr) == tolower(*sptr))
      {
         aptr++;
         sptr++;
      }
      if (*aptr == '/' && !*sptr)
      {
         /* buffer the in-configuration alias name */
         zptr = (sptr = AliasName) + sizeof(AliasName)-1;
         while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
         *sptr = '\0';
         /* buffer the in-configuration mapped user name */
         zptr = (sptr = UserName) + sizeof(UserName)-1;
         for (aptr++; *aptr && *aptr != '\n' && sptr < zptr; *sptr++ = *aptr++);
         *sptr = '\0';
         if (WatchEnabled)
            WatchThis (FI_LI, "ALIAS \"!AZ\" to \"!AZ\"", AliasName, UserName);
         *AliasNamePtrPtr = AliasName;
         *UserNamePtrPtr = UserName;
         return (TRUE);
      }
      while (*aptr && *aptr != '\n') aptr++;
   }

   if (WatchEnabled) WatchThis (FI_LI, "ALIAS \"!AZ\" NONE", *UserNamePtrPtr);
   return (TRUE);
}

/*****************************************************************************/
/*
Spawn a subprocess to execute an authentication agent.  The subprocess should
return a success code (any) to indicate valid credentials or an error code
(any) to indicate failure.  See description in module prologue.
*/ 

int LoginAgent
(
char *UserName,
char *Password,
char *RemoteAddr
)
{
   static $DESCRIPTOR (CommandFaoDsc, "!AZ \"!AZ\" \"!AZ\" \"!AZ\"");
   static $DESCRIPTOR (SysOutputDsc, "SYS$OUTPUT:");
   static $DESCRIPTOR (NullDeviceDsc, "NL:");

   int  status;
   unsigned long  ProcessId,
                  CompletionStatus;
   unsigned short  ShortLength;
   char  CommandBuffer [256];
   $DESCRIPTOR (CommandDsc, CommandBuffer);
   struct dsc$descriptor_s  *OutputDscPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginAgent()\n");

   status = sys$fao (&CommandFaoDsc, &ShortLength, &CommandDsc,
                     VWcmsConfig.LoginSource, UserName, Password, RemoteAddr);  
   if (VMSnok (status)) return (status);
   if (status == SS$_BUFFEROVF) return (SS$_RESULTOVF);
   CommandBuffer[ShortLength] = '\0';

   if (WatchEnabled)
   {
      fflush (stdout);
      OutputDscPtr = &SysOutputDsc;
   }
   else
      OutputDscPtr = &NullDeviceDsc;

   status = lib$spawn (&CommandDsc, 0, OutputDscPtr, 0, 0,
                       &ProcessId, &CompletionStatus, 0, 0, 0, 0, 0);
   CommandDsc.dsc$w_length = ShortLength;

   if (VMSnok (status)) return (status);

   if (WatchEnabled)
      WatchThis (FI_LI, "!AZ %X!8XL", VWcmsConfig.LoginSource, CompletionStatus);

   return (CompletionStatus);
}

/*****************************************************************************/
/*
Generate a string comprising the null-terminated elements of the session and
set a cookie containing it.
*/ 

void LoginSetSession
(
char *UserName,
char *RemoteAddr
)
{
   static $DESCRIPTOR (CredFaoDsc, "!UL\0!UL\0!AZ\0!AZ\0!AZ\0");

   int  status;
   unsigned short  CredLength;
   unsigned long  TimeSeconds;
   char  CredBuffer [256];
   $DESCRIPTOR (CredBufferDsc, CredBuffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginSetSession()\n");

   time (&TimeSeconds);

   /* the leading number just adds a little obfuscation to the mix */
   status = sys$fao (&CredFaoDsc, &CredLength, &CredBufferDsc,
                     TimeSeconds & 0x7f,
                     TimeSeconds + VWcmsConfig.LoginIdleSeconds,
                     RemoteAddr, UserName,
                     VWcmsConfig.LoginSecret);
   if (VMSnok(status)) ErrorExit (status, FI_LI);
   CredBuffer[0] = LOGIN_COOKIE_VERSION;

   LoginSetSessionCookie (CredBuffer, CredLength);
}

/*****************************************************************************/
/*
Get the session cookie if present.  Parse out the value, decode the base64 and
then decrypt the content.  Store that plain content in a static buffer,
allocate a dynamic string and return the location of that and it's length in
the supplied parameters.  Return the length of cookie or zero if there is none
(or an error processing it).
*/ 

char* LoginGetSessionCookie (int *CookieValueLengthPtr)

{
   int  CookieValueLength;
   char  *CookieValuePtr;
   rc4_key  rc4Key;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginGetSessionCookie()\n");

   if (CookieValueLengthPtr) *CookieValueLengthPtr = 0;

   if (!VWcmsConfig.LoginSecret) return (NULL);

   CookieValuePtr = CookieGet (LOGIN_SESSION_COOKIE);
   if (!CookieValuePtr) return (NULL);
   if (WatchEnabled)
      WatchThis (FI_LI, "!AZ=!AZ", LOGIN_SESSION_COOKIE, CookieValuePtr);

   /* decode from base64 to encrypted bytes */
   CookieValueLength = Base64Decode (CookieValuePtr);
   if (!CookieValueLength)
   {
      if (WatchEnabled) WatchThis (FI_LI, "BASE64 decode failed");
      return (NULL);
   }

   /* decrypt the RC4 encrypted credentials to plain bytes */
   rc4_prepare_key ((unsigned char*)VWcmsConfig.LoginSecret,
                    strlen(VWcmsConfig.LoginSecret), &rc4Key);
   rc4_crypt ((unsigned char*)CookieValuePtr, CookieValueLength, &rc4Key);

   if (CookieValueLengthPtr) *CookieValueLengthPtr = CookieValueLength;

   return (CookieValuePtr);
}

/*****************************************************************************/
/*
Create an field suitable for inclusion in an HTTP response to set the supplied
cookie value.  The cookie value is RC4 encrypted and the base64-encoded into
text suitable for use as a cookie value.  No expiry is set so the cookie should
disappear with the close of the browser.  A path and domain is supplied so the
cookie should only be be sent to VWCMS.
*/ 

void LoginSetSessionCookie
(
char *CookieContent,
int CookieLength
)
{
   int  status;
   char  *cptr, *sptr, *zptr,
         *Base64Ptr,
         *BufferPtr;
   rc4_key  rc4Key;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LoginSetSessionCookie()\n");

   if (!VWcmsConfig.LoginSecret) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (CookieLength < 0) CookieLength = strlen(CookieContent);
   BufferPtr = CgiLibVeeMemCalloc (CookieLength);
   if (!BufferPtr) ErrorExit (vaxc$errno, FI_LI);
   memcpy (BufferPtr, CookieContent, CookieLength);

   /* encrypt the plain credentials using RC4 */
   rc4_prepare_key ((unsigned char*)VWcmsConfig.LoginSecret,
                    strlen(VWcmsConfig.LoginSecret), &rc4Key);
   rc4_crypt ((unsigned char*)BufferPtr, CookieLength, &rc4Key);

   Base64Ptr = Base64Encode (BufferPtr, CookieLength);
   if (!Base64Ptr) ErrorExit (SS$_BUGCHECK, FI_LI);

   CookieSet (LOGIN_SESSION_COOKIE, Base64Ptr);
}

/*****************************************************************************/

